#!/usr/bin/perl
use strict;
use LWP::UserAgent;

##########################################################################
### "our" variables are configurable with the Configuration File
##########################################################################

# $AOpath = full path to AO installation
our $AOpath = "C:\\Program Files\\Anarchy Online\\";

# $GSPpath = full path to GridStream Player installation
our $GSPpath = "C:\\Program Files\\GridStream\\";

# $LyricsPath = full path to directory for static lyrics files
# If this directory exists, it will be checked for files matching:
# $artist\$song.txt
# $artist - $song.txt
# $song - $artist.txt
# $song.txt
# The first one of these found will be used instead of looking the
# song up on any lyrics search sites.
our $LyricsPath = $GSPpath."lyrics\\";

# $SetBackground = whether or not to set the desktop background
# Only the selection of background image is changed, not the tiling
# or background color settings.
our $SetBackground = 1;

# $ImagesPath = full path to desktop backgrounds
# Create a subdirectory here for each DJ you wish to have background
# images for.  One file will be chosen at random from those ending
# with the $ImagesTypes glob.  If "Playlist" is in the filename, it
# will only be chosen if the stream is determined to be on playlist.
# If no files have "Playlist" in their names, playlist mode will use
# one of the other files instead.
our $ImagesPath = $GSPpath."images\\";

# $ImageTypes = shell glob pattern to match the end of files that can be
# set as backgrounds
# ".bmp"  is pretty much your only option without using Active Desktop
our $ImageTypes = ".bmp";

# $Default = directory for default background images to use if the DJ name
# cannot be parsed
our $Default = "GSP";

# $AgentString = HTTP Agent string, in case one of the lyrics sites cares
our $AgentString = 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.0.4) Gecko/2008102920 Firefox/3.0.4';

# %Alias = hash of stream title regular expressions and the DJs / image
# directories they correspond to
# if ($SetBackground) stream titles will be parsed for the show name and
# matched against the directory names in $LyricsPath.
# The keys of this hash are inserted into the regular expression 
#   /(\b|DJ)($_)\b/i
# and if matched, the corresponding value is used as the directory name
# to choose a desktop background
our %Alias = (
	"LTK" => "Tarryk",
	"Sami" => "Samalicious",
	"Chai" => "Chaimera",
	"Synn" => "Synnthetic",
	"Anger Management" => "Ashval",
    "Lyk" => "Lykeios"
);

# %Sites = hash of site names and the details of how to search them
# ("site" => {
#   'referrer' => string for HTTP "Referer" header
#   'method' => "GET" or "POST"
#   'searchURI' => URL to search
#   'searchtext' => either a string or a hash
#      the string or the values will be interpolated, so the strings
#      '${song}' and '${artist}' can be used to insert those variables
#      into the search.  They are then URLencoded for the request
#      Use a hash for POST requests
#      Get can use either a hash or a preformatted string to append
#      to the searchURI
#   'regex' => regular expression to match the link to the result
#      use the named capturing subpatterns "link" and "match" to
#      capture the link and the artist/title it links to.  e.g.
#      /<a\s+href=\"\/lyrics\/(?<link>[^\"]+)\"\s*>(?<match>[^<]+)<\/a>/
#   'linkURI' => string to prepend to the matched (?<link>), in case 
#     it's relative
#   'exactstring' => what the matched (?<match>) should be, after all
#     HTML tags are stripped out, if it is an exact match
#   'altsearchURI' => if the first search doesn't match anything, try
#     this one, using the same regex, linkURI, and exactstring
#   'altsearchtext' => string or hash of variable(s) to use with the
#     altsearchURI, in the same format as 'searchtext'
#  }
# )
our %Sites = (
    "AbsoluteLyrics" => {
	    'referrer'      => "http://www.absolutelyrics.com/lyrics/searchpage/",
	    'method'        => "GET",
        'searchURI'     => "http://www.absolutelyrics.com/lyrics/search/?type=as&op=and&sort=s&q=",
	    'searchtext'    => '${song} - ${artist}',
    	'regex'         => '<li>\s*<a\s+href=\"\/lyrics\/view\/(?<link>[^\"]+)\"\s+title=\"(?<match>[^\"]*)\"',
	    'linkURI'       => "http://www.absolutelyrics.com/lyrics/view/*",
	    'exactstring'   => '$artist - $song lyrics',
	    'altsearchURI'  => "http://www.absolutelyrics.com/lyrics/search/?type=s&op=and&sort=s&q=",
	    'altsearchtext' => '${song}'
        } ,
    "MetroLyrics" => {
	    'referrer'      => "www.metrolyrics.com",
	    'method'        => "GET",
        'searchURI'     => "http://www.metrolyrics.com/search.php?category=artisttitle&action=Search&search=", 
	    'searchtext'    => '${song} - ${artist}',
    	'regex'         => '<td\s+class=\"First\"\s*>\s*<a\s+href=\"http:\/\/www.metrolyrics.com\/(?<link>[^\"]+)\">(?<match>[^<]+ Lyrics<\/a><\/td>\s*<td class=\"Last\"><a href=\"http:\/\/www.metrolyrics.com\/([^\"]+)\">[^<]+ Lyrics)<\/a><\/td>',
	    'linkURI'       => "http://www.metrolyrics.com/*",
	    'exactstring'   => '${song} Lyrics ${artist} Lyrics', 
	    'altsearchURI'  => "http://www.metrolyrics.com/search.php?category=title&action=Search&search=",
	    'altsearchtext' => '${song}'
        } ,
    "LyricsMode" => {
	    'referrer'      => "www.lyricsmode.com",
	    'method'        => "GET",
        'searchURI'     => "http://www.lyricsmode.com/search.php?what=songs&",
	    'searchtext'    => { 's' => '${song}', 'ss' => '${artist}' } ,
    	'regex'         => '<font\s+class=\"sm2\"\s*>\s*\d+\s*\.\s*<\/font>\s*&nbsp;\s*<a\s+href=\"\/lyrics\/[^\/]\/[^\/]+\/?\">(?<match>[^<]+<\/a><\/td>\s*<td>\s*<a\s+href=\"\/(?<link>[^\"]+)\"\s+class=\"b\"\s*>\s*[^<]+ lyrics)\s*<\/a>',
	    'linkURI'       => "http://www.lyricsmode.com/*",
	    'exactstring'   => '${artist} ${song} lyrics',
	    'altsearchURI'  => "http://www.lyricsmode.com/search.php?what=songs&s=",
	    'altsearchtext' => '${song}'
        }
# A-Z Lyrics stopped responding. Maybe they didn't like the repetitive searches
# and blocked the searches.  Or maybe they just had technical difficulties.
# Maybe I'll try them again later.
#    "A-Z Lyrics" => {
#	    'referrer'      => "http://www.azlyrics.com",
#	    'method'        => "GET", 
#       'searchURI'     => "http://search.azlyrics.com/cgi-bin/azseek.cgi?q=", 
#	    'searchtext'    => '${song} - ${artist}',
#    	'regex'         => '<FONT Face=Verdana size=2><a href=\"http:\/\/www.azlyrics.com\/lyrics\/(?<link>[^\"]+)\" TARGET=\"_blank\">(?<match>(?:[^<]|<[^\/]|<\/[^a]|<\/a[^>])+)<\/a><\/FONT>',
#	    'linkURI'       => "http://www.azlyrics.com/lyrics/*",
#	    'exactstring'   => '${artist} - ${song} lyrics',
#	    'altsearchURI'  => "http://search.azlyrics.com/cgi-bin/azseek.cgi?q=", 
#	    'altsearchtext' => '${song}'
#        }
);

# %ScriptOutput = hash of output script filenames and the strings to 
# interpolate (like 'searchtext' above) and print to them.
# These can then be invoked by the usual /script <filename> or just 
# /<filename>
our %ScriptOutput = (
    "lyrics"      => '/me sings along with \"$song\" by $artist.  <a href=\"text://<font color=CCInfoText>Artist: <font color=CCGreen>$artist</font><br>Song: <font color=CCGreen>$song</font><br>$HTMLresult</font>\">[lyrics]</a>\n',
    "checklyrics" => q[/text Lyrics <a href=\"text://<font color=CCInfoText>Artist: <font color=CCGreen>$artist</font><br>Song: <font color=CCGreen>$song</font><br>$HTMLresult</font>\">Results</a>: ".($lyricsfile?"file found":"$hits hit".($hits==1?"":"s"))." for \"$song\" by '$artist' :$checksummary\n],
	"llyrics"     => '/me looks up the <a href=\"text://<font color=CCInfoText>Artist: <font color=CCGreen>$artist</font><br>Song: <font color=CCGreen>$song</font><br>$HTMLresult</font>\">lyrics</a> to \"$song\" by $artist to read along.\n',
    );

# %InfoOutput = hash of infotext filenames and the strings to
# interpolate (like 'searchtext' above) and print to them.
# These can then be displayed with the /showinfo <filename>
# command.
our %InfoOutput = (
    "CheckLyrics.html" => '<font color=CCInfoText>Artist: <font color=CCGreen>$artist</font>\nSong: <font color=CCGreen>$song</font>\n$TEXTresult</font>\n'
    );

print "Variables initialized.  Loading configuration file.\n";

##########################################################################
### Configuration File
### Can be used to override or add to the default values of any of the
### configuration variables.  It is processed as normal perl, so just set 
### or modify the variables, like so:
### $AOpath = "D:\\Games\\Anarchy Online";
### $Alias{'Doc'} = "Klinical";
##########################################################################
my $ConfigFile = 'GSPlyrics_cfg.pl';
if (! -f $ConfigFile) {
    print STDERR "Configuration file '$ConfigFile' not found\n";
} elsif (! -r $ConfigFile) {
    print STDERR "Configuration file '$ConfigFile' not readable\n";
} else {
    my $rc = do($ConfigFile);

    if ($@) {
        print STDERR "ERROR: Failure compiling '$ConfigFile' - $@\n";
    } elsif (! defined($rc)) {
        print STDERR "ERROR: Failure reading '$ConfigFile' - $!\n";
    } elsif (! $rc) {
        print STDERR "ERROR: Failure processing '$ConfigFile'\n";
    }
}

print "Configuration finished.\n";

##########################################################################
### ensure all path variables end with a slash
##########################################################################
$AOpath =~ s/\\?$/\\/;
$GSPpath =~ s/\\?$/\\/;
$LyricsPath =~ s/\\?$/\\/;
$ImagesPath =~ s/\\?$/\\/;

##########################################################################
### These should not need to be changed - they're not optional
##########################################################################
# $ScriptPath = path to AO Scripts directory
my $ScriptPath = $AOpath."Scripts\\";
# $InfoPath = path to AO help text files
my $InfoPath = $AOpath."cd_image\\text\\help\\";
# $GSPscript = filename of the script GSPplayer outputs
my $GSPscript = $ScriptPath."gsp";

##########################################################################
### That's it for configuration
### The rest of the variables are for internal use, parsing and whatnot.
##########################################################################
my $song = "undefined";
my $artist = "undefined";
my $show = "undefined";
my $dj = "undefined";
my $lastdj = "undefined";
my $playlist = 0;
my $query;

my $lyricsfile = 0;
my $last = 0;
my $hits = 0;

my @stat;
my $changed;

my $searchURI;
my $agent = LWP::UserAgent->new;
$agent->env_proxy;
$agent->agent($AgentString);
my $response;
my @hits;
my %matches;
my %scores;
my $scorestring;
my $match;
my $script;
my $info;

my $syspinf;

if ($SetBackground) {
    use Win32::API;
    use constant SPI_SETDESKWALLPAPER  => 20;
    use constant SPIF_UPDATEANDSENDINI => 3;
    use constant NULL                  => 0;

    $syspinf = Win32::API->new('user32','SystemParametersInfo', ['I','I','P','I'], 'I')
        or warn "Could not import SystemParametersInfo function.\n";
}

chdir $ImagesPath;

##########################################################################
### main loop continues until SIGINT by Ctrl-C or killed by Alt-F4
##########################################################################
do {
    if (! -f $GSPscript) {
        die ("Cannot find '$GSPscript'\n")
    }
    @stat = stat $GSPscript;
	$changed = $stat[9];
	if ($changed != $last) {
        print "\n";
		$last = $changed;
		if (! open GSP, $GSPscript) {
			print "Could not open GSP Player info script '$GSPscript'";
		} else {
			$song = "unknown";
			$artist = "unknown";
			$show = "unknown";
			$dj = $Default;
			%matches = ();
			%scores = ();
			my $HTMLresult = "";
			my $TEXTresult = "";
			my $checksummary = "";

			while (<GSP>) {
				chomp;
				if (/Listening<\/a> to <font color=CCWhite>\s*(.*\S)\s+-\s+(.*\S)\s*<\/font>\s*(.*\S)/) {
					$artist = $1;
					$song = $2;
					$show = $3;
                    if ($SetBackground) {
    					foreach my $f (glob "*\\") {
	    					if ($f =~ /([^\\]*)\\$/) {
		    					if ($show =~ /(\b|DJ)($1)\b/i) {
			    					$dj = $2;
				    			}
					    	}
					    }
					    if ($dj eq $Default) {
    						print "No images directory matching '$show'\n";
						    my $alias;
						    foreach $alias (keys(%Alias)) {
    							if ($show =~ /(\b|DJ)($alias)\b/i) {
								    $dj = $Alias{$alias};
							    }
						    }
					    }
					    $playlist = ($show =~ /playlist/i);
					    print "Current DJ: $dj".($playlist? " (playlist)" : "")."\n";
					    if ($dj ne $lastdj && -d "$dj") {
    						my @images = glob "$dj\\*$ImageTypes";
						    if (!$playlist) {@images = grep(!/playlist/i, @images);}
						    elsif (my @playlist=grep(/playlist/i, @images)) {@images = @playlist;}
						    if ($#images >= 0) {
    							$lastdj = $dj;
							    my $image = $images[int(rand($#images+1))];
							    $syspinf->Call(SPI_SETDESKWALLPAPER, 0, $ImagesPath.$image, SPIF_UPDATEANDSENDINI);
						    }
                        }
					}
					print "Listening to '$song' by $artist, $show\n";
				} elsif (/Listening<\/a> to <font color=CCWhite>\s*TALK LIVE\s*<\/font>\s*(.*\S)/i) {
					print "Listening to Talk Live, $1\n";
				} elsif (/./) {
					print "$_\n";
				}
			}
			close GSP;

			if (($song ne "unknown" || $artist ne "unknown" ||  $show ne "unknown")
				&& $artist ne "SWEEP" && $artist ne "PSA" && $artist ne "PSA - GSP Disclaimer"){
				$scorestring = "$artist $song";
				$scorestring =~ s/\s+/ /g;
				$scorestring =~ s/([^-a-z0-9_!@#%&;:<>, ])/\\$1/ig;
				$scorestring = '(\b'.join('\b|\b', split(/\s/, $scorestring)).'\b)';
		
				$hits = 0;
				$lyricsfile = 0;

				if (-d $LyricsPath) {
					if (-r _) {
						if (-d "$LyricsPath\\$artist") {
							if (-f "$LyricsPath\\$artist\\$song.txt") {
								$lyricsfile = "$LyricsPath\\$artist\\$song.txt";
							}
						}
						if (!$lyricsfile && -f "$LyricsPath\\$artist - $song.txt") {
							$lyricsfile = "$LyricsPath\\$artist - $song.txt";
						}
						if (!$lyricsfile && -f "$LyricsPath\\$song - $artist.txt") {
							$lyricsfile = "$LyricsPath\\$song - $artist.txt";
						}
						if (!$lyricsfile && -f "$LyricsPath\\$song.txt") {
							$lyricsfile = "$LyricsPath\\$song.txt";
						}
						if ($lyricsfile) {
							open LYRICS, $lyricsfile;
							while (<LYRICS>) {
								s/[\r\n]+//g;
								$HTMLresult .= "$_<br>";
								$TEXTresult .= "$_\n";
							}
							close LYRICS;
						} else {
                            my $site;
				            foreach $site (keys(%Sites)) {
                                my $exactstring = eval('"'.$Sites{$site}{'exactstring'}.'"');
                                my $searchtext;
                                my $altsearchtext;
                                if (ref($Sites{$site}{'searchtext'})) {
                                    $searchtext = {};
                                    foreach ($Sites{$site}{'searchtext'}) {
                                        $searchtext->{$_}=eval('"'.$Sites{$site}{'searchtext'}{$_}.'"')
                                    }
                                } else {
                                    $searchtext = eval($Sites{$site}{'searchtext'});
                                }
                                if (ref($Sites{$site}{'altsearchtext'})) {
                                    $altsearchtext = {};
                                    foreach ($Sites{$site}{'altsearchtext'}) {
                                        $altsearchtext->{$_}=eval('"'.$Sites{$site}{'altsearchtext'}{$_}.'"')
                                    }
                                } else {
                                    $altsearchtext = eval('"'.$Sites{$site}{'altsearchtext'}.'"');
                                }
							    search ($site,
    								    $Sites{$site}{'referrer'},
    								    $Sites{$site}{'method'},
								        $Sites{$site}{'searchURI'},
								        $searchtext,
								        $Sites{$site}{'regex'},
								        $Sites{$site}{'linkURI'},
								        $exactstring,
								        $scorestring,
								        $Sites{$site}{'altsearchURI'},
								        $altsearchtext
								        );
                            }

							@hits = ();
							my $lastscore = 1000000;
							my $score = 0;
							my $site = "";
							foreach $site (sort { 0 - ($scores{$a} <=> $scores{$b}) } keys %scores) {
								$score = $scores{$site};
#								print "site $site , score $score , lastscore $lastscore\n";
								if ($score < $lastscore) {
									if ($score == 100) {
										$HTMLresult .= "<br>Lyrics found on:<br>";
										$TEXTresult .= "\nLyrics found on:\n";
									} elsif ($score == 0) {
										$HTMLresult .= "<br>No match on:<br>";
										$TEXTresult .= "\nNo match on:\n";
									} elsif ($lastscore >= 100) {
										$HTMLresult .= "<br>Possible match on:<br>";
										$TEXTresult .= "\nPossible match on:\n";
									}
									$lastscore = $score;
								}
								if ($score == 100) {
									$checksummary .= " +".$site;
								} elsif ($score == 0) {
									$checksummary .= " -".$site;
								} else {
									$checksummary .= " ?".$site;
								} 
								$HTMLresult .= $matches{$site};
								$TEXTresult .= $matches{$site};
								push @hits, $site if $score > 0;
							}

							if ($hits) {
								print "$hits hit".($hits==1?"":"s")." for '$song' by '$artist' : ".join(", ",@hits)."\n";
							} else {
								$HTMLresult .= "None found.  Sorry.<br>";
								$TEXTresult .= "None found.  Sorry.\n";
								print "No matches found.\n";
							}
		
							$HTMLresult .= "<br><br><a href='chatcmd:///start http://www.google.com/search?q=".aoencode("\"$artist\" +\"$song\" +lyrics")."'>Google it</a><br>";
							$TEXTresult .= "\n\n<a href='chatcmd:///start http://www.google.com/search?q=".aoencode("\"$artist\" +\"$song\" +lyrics")."'>Google it</a>\n";
		
						}
					} else {
						print "Directory '$LyricsPath' not readable\n";
					}
				} else {
					print "Directory '$LyricsPath' not found\n";
				}
                
                foreach $script (keys %ScriptOutput) {
                    open SCRIPT, ">$ScriptPath$script";
    				print SCRIPT eval('"'.$ScriptOutput{$script}.'"');
	    			close SCRIPT;
                }

                foreach $info (keys %InfoOutput) {
                    open INFO, ">$InfoPath$info";
    				print INFO eval('"'.$InfoOutput{$info}.'"');
	    			close INFO;
                }
			}
		}
	}
	sleep 1;
} while (1);

sub urlencode {
	my $str = shift;
	$str =~ s/([^A-Za-z0-9 ])/sprintf("%%%02X", ord($1))/eg;
	$str =~ s/ /+/g;
	return $str;
}

sub aoencode {
	my $str = shift;
	$str =~ s/([^A-Za-z0-9 ])/sprintf("%%%%9%02X", ord($1))/eg;
	$str =~ s/ /+/g;
	return $str;
}

sub search {
	my $site = shift;
	my $referrer = shift;
	my $method = shift;
	my $searchURI = shift;
	my $searchtext = shift;
	my $regex = shift;
	my $linkURI = shift;
	my $exactstring = shift;
	my $scorestring = shift;
	my $altsearchURI = shift;
	my $altsearchtext = shift;
	$exactstring =~ s/\s+/ /g;
	$exactstring =~ s/([^-a-z0-9_!@#%&;:<>, ])/\\$1/ig;
	$exactstring = "(The )?$exactstring";
	$exactstring =~ s/\s+/\\s+/g;
				
	if ($method =~ /get/i) {
		my $request;
		if (ref($searchtext)) {
			$request = $searchURI.join("&", map($_."=".urlencode($$searchtext{$_}), keys(%$searchtext)));
		} else {
			$request = $searchURI.urlencode($searchtext);
		}
#		print "Searching: $request\n";
		$response = $agent->get($request, "Referer: $referrer");
	} elsif ($method =~ /post/i) {
		my %searchhash = %$searchtext;
#		print "Searching: $searchURI";
#		my $i=0;
#		my $k;
#		foreach $k (keys %searchhash) {
#			print "".(($i++)?"&":"?")."$k=".urlencode($searchhash{$k});
#		}
#		print "\n";
		$response = $agent->post($searchURI,%searchhash, "Referer: $referrer");				
	} else {
		die "Unknown Method: $method\n";
	}
	
	if ($response->is_success) {
		if ($response->content =~ /$regex/i) {
			my $link = $+{'link'};
			my $match = $+{'match'};
			$match =~ s/<[^>]*>/ /ig;
			$match =~ s/\s+/ /ig;
			$hits++;
			$linkURI =~ s/\*/$link/;
			$matches{$site} = "*<a href='chatcmd:///start $linkURI'>$site</a><br>";
			if ($match =~ /^\s*$exactstring\s*$/i) {
				$scores{$site} = 100;
			} else {
				$scores{$site} = 0+ (()=$match=~/$scorestring/ig);
			}
			print "$site '$match' score: ".$scores{$site}."\n";
		} elsif ($altsearchURI) {
			if ($method =~ /get/i) {
				my $request;
				if (ref($altsearchtext)) {
					$request = $altsearchURI.join("&", map($_."=".urlencode($$altsearchtext{$_}), keys(%$altsearchtext)));
				} else {
					$request = $altsearchURI.urlencode($altsearchtext);
				}
#				print "Searching: $request\n";
				$response = $agent->get($request, "Referer: $referrer");
			} elsif ($method =~ /post/i) {
				my %searchhash = %{$altsearchtext};
				print "Searching: $altsearchURI";
				my $i=0;
				foreach (keys(%searchhash)) {
					print "".(($i++)?"&":"?").$_."=".urlencode($searchhash{$_});
				}
				print "\n";
				$response = $agent->post($altsearchURI,%searchhash, "Referer: $referrer");				
			} else {
				die "Unknown Method: $method\n";
			}
			if ($response->is_success) {
				if ($response->content =~ /$regex/i) {
					my $link = $+{'link'};
					my $match = $+{'match'};
					$match =~ s/<[^>]*>/ /ig;
					$match =~ s/\s+/ /ig;
					$hits++;
					$linkURI =~ s/\*/$link/;
					$matches{$site} = "*<a href='chatcmd:///start $linkURI'>$site</a><br>";
					if ($match =~ /^\s*$exactstring\s*$/i) {
						$scores{$site} = 100;
					} else {
						$scores{$site} = 0+ (()=$match=~/$scorestring/ig);
					}
					print "$site '$match' score: ".$scores{$site}."\n";
				} else {
					if ($method =~ /get/i) {
						if (ref($altsearchtext)) {
							$query = join("&", map($_."=".aoencode($$altsearchtext{$_}), keys(%$altsearchtext)));
						} else {
							$query = aoencode($altsearchtext);
						}
					} else {
						$query = "";
					}
					$matches{$site} = "-<a href='chatcmd:///start $altsearchURI$query'>$site</a><br>";
					$scores{$site} = 0;
				}
			} else {
				print $response->status_line . "\n";
			}
		} else {
			if ($method =~ /get/i) {
				if (ref($altsearchtext)) {
					$query = join("&", map($_."=".aoencode($$searchtext{$_}), keys(%$searchtext)));
				} else {
					$query = aoencode($searchtext);
				}
			} else {
				$query = "";
			}
			$matches{$site} = "-<a href='chatcmd:///start $searchURI$query'>$site</a><br>";
			$scores{$site} = 0;
		}
	} else {
		print $site.": ".$response->status_line . "\n";
	}
}
